@@ -1,6 +1,5 @@
 from enum import Enum
 from typing import Optional
-import os
 
 REGION = 'us-east-1'
 NAMESPACE = 'kube-system'
@@ -16,15 +15,31 @@ KEDA_NAMESPACE = 'keda'
 KEDA_SERVICE_ACCOUNT_NAME = 'keda-operator'
 KEDA_IAM_ROLE_NAME = 'keda-iam-role'
 
+CREATED_BY_TAG = 'CreatedBy'
+CLUSTER_NAME_TAG = 'ClusterName'
+TENSORFUSE_STRING = 'tensorfuse'
+
+
 TENSORFUSE_NAMESPACES = ['amazon-cloudwatch', 'kube-system', 'istio-system', 'knative-serving', 'kube-node-lease',
-    'kube-public', 'kube-system', ]
+                         'kube-public', 'kube-system', 'nydus-snapshotter' ]
+
+LOCKED_AWS_CLI_VERSION = "2.22.8"
+LOCKED_EKSCTL_VERSION = "0.194.0"
+LOCKED_KUBECTL_VERSION = "1.29.11"
+LOCKED_HELM_VERSION = "3.16.3"
+LOCKED_ISTIO_VERSION = "1.24.0"
+
+
+NODEPOOL = 'nodepool'
+EC2NODECLASS = 'ec2nodeclass'
+GPU_NODEPOOL_VERSION = 'gpu-v1'
 
 
 def get_cluster_name():
     return "tensorkube"
 
 
-def get_mount_policy_name(cluster_name:str, env: Optional[str] = None) -> str:
+def get_mount_policy_name(cluster_name: str, env: Optional[str] = None) -> str:
     if env:
         return f'{cluster_name}-mountpoint-policy-env-{env}'
     return f'{cluster_name}-mountpoint-policy'
@@ -35,10 +50,17 @@ def get_mount_driver_role_name(cluster_name):
 
 
 class Events(Enum):
+    LOGGED_IN = 'logged-in'
+
+    COMMAND_RUN = 'command-run'
+
     CONFIGURE_START = 'configure-start'
     CONFIGURE_END = 'configure-end'
     CONFIGURE_ERROR = 'configure-error'
 
+    TEARDOWN_START = 'teardown-start'
+    TEARDOWN_END = 'teardown-end'
+
     DEPLOY_START = 'deploy-start'
     DEPLOY_END = 'deploy-end'
     DEPLOY_ERROR = 'deploy-error'
@@ -47,8 +69,13 @@ class Events(Enum):
     DEVCONTAINER_START = 'devcontainer-start'
     DEVCONTAINER_STOP = 'devcontainer-stop'
     DEVCONTAINER_LIST = 'devcontainer-list'
+    DEVCONTAINER_DELETE = 'devcontainer-delete'
     DEVCONTAINER_ERROR = 'devcontainer-error'
 
+    JOB_DEPLOY_START = 'job-deploy-start'
+    JOB_DEPLOY_END = 'job-deploy-end'
+    JOB_DEPLOY_ERROR = 'job-deploy-error'
+
     TEST_START = 'test-start'
     TEST_END = 'test-end'
 
