@@ -1,12 +1,17 @@
+import json
 import os
 import platform
 import subprocess
 from string import Template
 
+import click
+import semver
 from pkg_resources import resource_filename
 
-from tensorkube.services.aws_service import get_aws_account_id, get_karpenter_namespace
+from tensorkube.constants import CliColors, LOCKED_EKSCTL_VERSION
 from tensorkube.constants import get_cluster_name
+from tensorkube.services.aws_service import get_aws_account_id, get_karpenter_namespace
+from tensorkube.services.error import CLIVersionError
 
 
 # create base cluster using eksctl
@@ -51,25 +56,27 @@ def delete_cluster():
 def check_and_install_eksctl():
     """Check if eksctl is installed and if not isntall it."""
     try:
-        subprocess.run(["eksctl", "version"], check=True)
+        version_cmd = subprocess.run(["eksctl", "version", "-o", "json"], check=True, stdout=subprocess.PIPE)
+        version_dict = json.loads(version_cmd.stdout.decode("utf-8"))
+        version = semver.VersionInfo.parse(version_dict["Version"])
+        locked_eksctl_version = semver.VersionInfo.parse(LOCKED_EKSCTL_VERSION)
+        # check if the eksctl version is above locked eksctl version
+        print("eksctl is already installed.")
+        if (version.major, version.minor) < (locked_eksctl_version.major, locked_eksctl_version.minor):
+            text = f"eksctl version is {version}. Please upgrade eksctl to version above {LOCKED_EKSCTL_VERSION}."
+            click.echo(click.style(text, bold=True, fg=CliColors.ERROR.value))
+            raise CLIVersionError(text)
     except Exception as e:
+        if isinstance(e, CLIVersionError):
+            raise e
+        click.echo(
+            click.style("Eksctl not found. Proceeding with installation. Might require sudo password.", bold=True,
+                        fg=CliColors.WARNING.value))
         # check if the operating system is mac and install eksctl
-        if platform.system() == "Darwin":
+        if platform.system() == "Darwin" or platform.system() == "Linux":
             try:
-                subprocess.run(["brew", "install", "eksctl"])
-            except Exception as mac_e:
-                print("Unable to install eksctl. Please install eksctl manually.")
-                raise mac_e
-        elif platform.system() == "Linux":
-            try:
-                # curl --silent --location "https://github.com/weaveworks/eksctl/releases/latest/download/eksctl_$(uname -s)_amd64.tar.gz" | tar xz -C /tmp
-                # sudo mv /tmp/eksctl /usr/local/bin
-                # eksctl version
-                download_command = [
-                    "curl", "--silent", "--location",
-                    "https://github.com/weaveworks/eksctl/releases/latest/download/eksctl_$(uname -s)_amd64.tar.gz",
-                    "-o", "/tmp/eksctl.tar.gz"
-                ]
+                loc = f"https://github.com/eksctl-io/eksctl/releases/download/v{LOCKED_EKSCTL_VERSION}/eksctl_Darwin_amd64.tar.gz"
+                download_command = ["curl", "--silent", "--location", loc, "-o", "/tmp/eksctl.tar.gz"]
                 subprocess.run(" ".join(download_command), shell=True, check=True)
                 # Extract the tar.gz file
                 extract_command = ["tar", "xzf", "/tmp/eksctl.tar.gz", "-C", "/tmp"]
@@ -78,9 +85,9 @@ def check_and_install_eksctl():
                 move_command = ["sudo", "mv", "/tmp/eksctl", "/usr/local/bin"]
                 subprocess.run(move_command, check=True)
                 subprocess.run(["eksctl", "version"])
-            except Exception as linux_e:
+            except Exception as e:
                 print("Unable to install eksctl. Please install eksctl manually.")
-                raise linux_e
+                raise e
         else:
             print("eksctl is not installed. Please install eksctl manually.")
             raise e
