@@ -1,7 +1,7 @@
 import subprocess
 import time
 from threading import Thread
-from typing import List, Optional, Dict
+from typing import List, Optional, Dict, Any
 
 import click
 import yaml
@@ -12,7 +12,8 @@ from pkg_resources import resource_filename
 from tensorkube.constants import NAMESPACE, REGION, PodStatus, BUILD_TOOL, DEFAULT_NAMESPACE, TENSORFUSE_NAMESPACES, \
     get_cluster_name, CliColors
 from tensorkube.services.aws_service import get_aws_account_id
-import json
+import boto3
+import os
 
 def get_s3_pv_name(env_name: Optional[str] = None):
     if env_name:
@@ -167,116 +168,9 @@ def create_build_pv_and_pvc(bucket_name: str, region: str = REGION, env: Optiona
             raise e
 
 
-
-def get_buildkit_command(sanitised_project_name: str, image_tag: str, image_url: Optional[str] =None, upload_to_nfs: bool= False, region: str = REGION):
-    aws_account_number =  get_aws_account_id()
-    command = [ "/bin/sh", "-c",
-        f"""
-        apk add --no-cache curl unzip aws-cli docker
-        aws ecr get-login-password --region {region} | docker login --username AWS --password-stdin {aws_account_number}.dkr.ecr.{region}.amazonaws.com
-        buildctl-daemonless.sh build\
-            --frontend dockerfile.v0\
-            --local context=/data/build/{sanitised_project_name}\
-            --local dockerfile=/data/build/{sanitised_project_name}\
-            --output type=image,name={image_url},push=true\
-        """
-    ]
-
-    if upload_to_nfs:
-        command =  ["/bin/sh", "-c", f"""
-        mkdir -p /data/cache
-        mkdir -p /test/tars/{sanitised_project_name} /mnt/efs/images/{sanitised_project_name}/{image_tag}/rootfs /test/tars/{sanitised_project_name} /test/images/{sanitised_project_name}/rootfs
-        time buildctl-daemonless.sh build\
-            --frontend dockerfile.v0\
-            --local context=/data/build/{sanitised_project_name}\
-            --local dockerfile=/data/build/{sanitised_project_name}\
-            --output type=tar,dest=/test/tars/{sanitised_project_name}/{sanitised_project_name}-{image_tag}.tar
-        if [ $? -ne 0 ]; then
-            echo "buildctl-daemonless.sh command failed"
-            exit 1
-        fi
-        echo "Extracting rootfs from tar file"
-        time tar -xf /test/tars/{sanitised_project_name}/{sanitised_project_name}-{image_tag}.tar -C /test/images/{sanitised_project_name}/rootfs --checkpoint=1000 --checkpoint-action=echo="Extracted #%u: %T"
-        if [ $? -ne 0 ]; then
-            echo "extracting rootfs from tar file failed"
-            exit 1
-        fi
-        echo "Extracted rootfs from tar file"
-        echo "Uploading rootfs to NFS"
-        cd /test/images/{sanitised_project_name}/rootfs
-        find . -type d | split -l 5000 - batch_dir_
-        find . -type f | split -l 5000 - batch_file_
-        find . -type l | split -l 5000 - batch_link_
-
-        total_batches=$(ls batch_* | wc -l)
-        processed_batches=0
-        total_time=0
-
-        process_batches() {{
-            local batch_type=$1
-            for batch_file in batch_${{batch_type}}_*; do
-                start_time=$(date +%s)
-                
-                if [ "$batch_type" == "dir" ]; then
-                    cat $batch_file | parallel -j 15 mkdir -p /mnt/efs/images/{sanitised_project_name}/{image_tag}/rootfs/{{}}
-                elif [ "$batch_type" == "file" ]; then
-                    cat $batch_file | parallel -j 15 cp --parents {{}} /mnt/efs/images/{sanitised_project_name}/{image_tag}/rootfs/
-                elif [ "$batch_type" == "link" ]; then
-                    cat $batch_file | parallel -j 15 cp --parents -P {{}} /mnt/efs/images/{sanitised_project_name}/{image_tag}/rootfs/
-                fi
-                
-                end_time=$(date +%s)
-                batch_time=$((end_time - start_time))
-                total_time=$((total_time + batch_time))
-                processed_batches=$((processed_batches + 1))
-                remaining_batches=$((total_batches - processed_batches))
-                avg_time_per_batch=$((total_time / processed_batches))
-                estimated_remaining_time=$((avg_time_per_batch * remaining_batches))
-                
-                echo "Processed batch $processed_batches/$total_batches. Remaining: $remaining_batches. Time for this batch: $batch_time seconds. Estimated remaining time: $estimated_remaining_time seconds."
-                rm $batch_file
-            done
-        }}
-
-        process_batches "dir"
-        process_batches "file"
-        process_batches "link"
-        echo "Uploaded rootfs to NFS" """]
-    return command
-
         
 
-def apply_k8s_buildkit_config(sanitised_project_name: str, image_tag: str, region: str = REGION,
-                              env_name: Optional[str] = None, context_name: Optional[str] = None, image_url: Optional[str] = None, upload_to_nfs: bool = False):
-    if not context_name:
-        context_name = get_tensorkube_cluster_context_name()
-        if not context_name:
-            return None
-    k8s_api_client = config.new_client_from_config(context=context_name)
 
-    buildkit_config_file_path = resource_filename('tensorkube', 'configurations/build_configs/buildkit.yaml')
-    with open(buildkit_config_file_path) as f:
-        buildkit_config = yaml.safe_load(f)
-
-    buildkit_config['metadata']['name'] = 'buildkit-{}'.format(sanitised_project_name)
-    buildkit_config['spec']['template']['spec']['containers'][0]['env'][0]['value'] = region
-
-    # Include the namespace in the buildkit configuration
-    # Replace 'default' with your default namespace if needed
-    namespace_to_use = env_name if env_name else DEFAULT_NAMESPACE
-    buildkit_config['metadata']['namespace'] = namespace_to_use
-
-    # Modify volume claim names based on env_name
-    for volume in buildkit_config['spec']['template']['spec']['volumes']:
-        if volume['name'] == 'persistent-storage':
-            volume['persistentVolumeClaim']['claimName'] = get_s3_claim_name(env_name=env_name)
-        elif volume['name'] == 'efs-pvc':
-            volume['persistentVolumeClaim']['claimName'] = get_efs_claim_name(env_name=env_name)
-
-    buildkit_config['spec']['template']['spec']['containers'][0]['command'] = get_buildkit_command(sanitised_project_name, image_tag, image_url, upload_to_nfs, region)
-
-    utils.create_from_dict(k8s_api_client, buildkit_config)
-    click.echo('Deployed a Buildkit image')
 
 
 def apply_image_cleanup_job(sanitised_project_name: str, image_tags: List[str], env: Optional[str] = None,
@@ -860,6 +754,16 @@ def list_all_namespaces(context_name: Optional[str] = None):
     return [namespace.metadata.name for namespace in namespaces]
 
 
+def ssh_into_pod(pod_name: str, namespace: str, context_name: Optional[str] = None):
+    if not context_name:
+        context_name = get_tensorkube_cluster_context_name()
+        if not context_name:
+            return None
+    
+    command = ["kubectl", "--context", f"{context_name}", "exec", "-it", pod_name, "-n", namespace, "-c",
+               "user-container", "--", "sh"]
+    subprocess.run(command)
+
 def ssh_into_pod_with_podman(pod_name: str, namespace: str, context_name: Optional[str] = None):
     if not context_name:
         context_name = get_tensorkube_cluster_context_name()
@@ -1193,3 +1097,368 @@ def delete_trigger_authentication(trigger_auth_name: str):
     except ApiException as e:
         print(f"Exception when calling CustomObjectsApi: {e}")
         return None
+
+def add_user_to_aws_auth_group(user_name, group_name):
+    v1 = client.CoreV1Api()
+    # details of the aws-auth ConfigMap
+    namespace = "kube-system"
+    configmap_name = "aws-auth"
+
+    configmap = v1.read_namespaced_config_map(configmap_name, namespace)
+    map_users = yaml.safe_load(configmap.data.get("mapUsers", "[]"))
+    # Append the new user to the mapUsers list
+    new_user = {
+        "userarn": f"arn:aws:iam::{get_aws_account_id()}:user/{user_name}",
+        "username": user_name,
+        "groups": [group_name]
+    }
+
+    map_users.append(new_user)
+    # Update the mapUsers back into the ConfigMap data
+    configmap.data["mapUsers"] = yaml.dump(map_users)
+
+    # Update the ConfigMap with the new user
+    v1.replace_namespaced_config_map(name=configmap_name, namespace=namespace, body=configmap)
+
+
+def add_user_to_k8s(user_name):
+    config.load_kube_config()
+    rbac_api = client.RbacAuthorizationV1Api()
+    cluster_role_name = f"{get_cluster_name()}-{user_name}-job-cluster-role"
+    cluster_role = client.V1ClusterRole(
+        metadata=client.V1ObjectMeta(name=cluster_role_name),
+        rules=[
+            client.V1PolicyRule(
+                api_groups=["keda.sh"],
+                resources=["scaledjobs"],
+                verbs=["create", "get", "list", "delete"]
+            )
+        ]
+    )
+    try:
+        rbac_api.create_cluster_role(body=cluster_role)
+    except ApiException as e:
+        if e.status == 409:
+            print(f"ClusterRole {cluster_role_name} already exists.")
+        else:
+            raise
+    print(f"ClusterRole {cluster_role_name} created successfully.")
+    group_name = f"{get_cluster_name()}-{user_name}-job-access"
+    cluster_role_binding_name = f"{get_cluster_name()}-{user_name}-job-cluster-role-binding"
+    # Define ClusterRoleBinding to bind the role to a user
+    cluster_role_binding = client.V1ClusterRoleBinding(
+        metadata=client.V1ObjectMeta(name=cluster_role_binding_name),
+        role_ref=client.V1RoleRef(
+            api_group="rbac.authorization.k8s.io",
+            kind="ClusterRole",
+            name=cluster_role_name
+        ),
+        subjects=[
+            client.RbacV1Subject(
+                kind="Group",
+                name=group_name,
+            )
+        ]
+    )
+    try:
+        rbac_api.create_cluster_role_binding(body=cluster_role_binding)
+    except ApiException as e:
+        if e.status == 409:
+            print(f"ClusterRoleBinding {cluster_role_binding_name} already exists.")
+        else:
+            raise
+    print(f"ClusterRoleBinding {cluster_role_binding_name} created successfully.")
+    add_user_to_aws_auth_group(user_name, group_name)
+    return
+
+
+
+def generate_kubeconfig():
+    eks_client = boto3.client(
+        'eks',
+        region_name=REGION,
+    )
+
+    cluster_name = get_cluster_name()
+    # Retrieve cluster information
+    cluster_info = eks_client.describe_cluster(name=get_cluster_name())['cluster']
+
+    # Extract cluster details
+    endpoint = cluster_info['endpoint']
+    cluster_ca_certificate = cluster_info['certificateAuthority']['data']
+
+
+    # Create a kubeconfig dictionary
+    kubeconfig_data = {
+        "apiVersion": "v1",
+        "clusters": [{
+            "cluster": {
+                "server": endpoint,
+                "certificate-authority-data": cluster_ca_certificate
+            },
+            "name": f'arn:aws:eks:us-east-1:{get_aws_account_id()}:cluster/{cluster_name}'
+        }],
+        "contexts": [{
+            "context": {
+                "cluster": f'arn:aws:eks:us-east-1:{get_aws_account_id()}:cluster/{cluster_name}',
+                "user": f'arn:aws:eks:us-east-1:{get_aws_account_id()}:cluster/{cluster_name}'
+            },
+            "name": f'arn:aws:eks:us-east-1:{get_aws_account_id()}:cluster/{cluster_name}'
+        }],
+        "current-context": f'arn:aws:eks:us-east-1:{get_aws_account_id()}:cluster/{cluster_name}',
+        "kind": "Config",
+        "users": [{
+            "name": f'arn:aws:eks:us-east-1:{get_aws_account_id()}:cluster/{cluster_name}',
+            "user": {
+                "exec": {
+                    "apiVersion": "client.authentication.k8s.io/v1beta1",
+                    "command": "aws",
+                    "args": [
+                        "eks",
+                        "get-token",
+                        "--cluster-name", cluster_name
+                    ],
+                    "env": [
+                        {"name": "AWS_DEFAULT_REGION", "value": REGION}
+                    ]
+                }
+            }
+        }]
+    }
+    kubeconfig_dir = os.path.expanduser('~/.kube')
+    kubeconfig_file = os.path.join(kubeconfig_dir, 'config')
+
+    # Create the directory if it does not exist
+    if not os.path.exists(kubeconfig_dir):
+        os.makedirs(kubeconfig_dir,exist_ok=True)
+
+    # Load existing kubeconfig or create a new one if it doesn't exist
+    if os.path.exists(kubeconfig_file):
+        with open(kubeconfig_file, 'r') as f:
+            existing_config = yaml.safe_load(f) or {}
+    else:
+        existing_config = {'apiVersion': 'v1', 'kind': 'Config', 'clusters': [], 'contexts': [], 'users': []}
+
+    # Merge new kubeconfig entry into existing config
+    # Function to update or add an entry in the list
+    def update_or_add_entry(existing_list, new_entry, key):
+        for i, entry in enumerate(existing_list):
+            if entry[key] == new_entry[key]:
+                existing_list[i] = new_entry  # Update the existing entry
+                return
+        existing_list.append(new_entry)  # Add as a new entry if not found
+
+    # Merge new kubeconfig entry into existing config
+    update_or_add_entry(existing_config['clusters'], kubeconfig_data['clusters'][0], 'name')
+    update_or_add_entry(existing_config['contexts'], kubeconfig_data['contexts'][0], 'name')
+    update_or_add_entry(existing_config['users'], kubeconfig_data['users'][0], 'name')
+    existing_config['current-context'] = kubeconfig_data['current-context']
+
+
+    # Write back to the default kubeconfig file
+    with open(kubeconfig_file, 'w') as f:
+        yaml.dump(existing_config, f)
+
+def create_service_account(name: str, namespace: str, context_name: Optional[str] = None):
+    if not context_name:
+        context_name = get_tensorkube_cluster_context_name()
+        if not context_name:
+            raise Exception("No context found for the cluster. Please configure Tensorfuse properly.")
+
+    k8s_client = config.new_client_from_config(context=context_name)
+    v1 = client.CoreV1Api(k8s_client)
+    service_account = client.V1ServiceAccount()
+    service_account.metadata = client.V1ObjectMeta(name=name)
+    try:
+        v1.create_namespaced_service_account(namespace=namespace, body=service_account)
+    except ApiException as e:
+        raise e
+    
+
+def create_service_account_with_role_arn(name: str, namespace: str, role_arn: str, context_name: Optional[str] = None):
+    if not context_name:
+        context_name = get_tensorkube_cluster_context_name()
+        if not context_name:
+            raise Exception("No context found for the cluster. Please configure Tensorfuse properly.")
+
+    k8s_client = config.new_client_from_config(context=context_name)
+    v1 = client.CoreV1Api(k8s_client)
+
+    # Check if the ServiceAccount already exists
+    try:
+        v1.read_namespaced_service_account(name=name, namespace=namespace)
+        print(f"ServiceAccount '{name}' already exists in namespace '{namespace}'.")
+    except ApiException as e:
+        if e.status == 404:
+            # ServiceAccount does not exist; create it
+            service_account = client.V1ServiceAccount(
+                metadata=client.V1ObjectMeta(
+                    name=name,
+                    annotations={"eks.amazonaws.com/role-arn": role_arn},
+                    namespace=namespace
+                ),
+            )
+            try:
+                v1.create_namespaced_service_account(namespace=namespace, body=service_account)
+                print(f"ServiceAccount '{name}' created in namespace '{namespace}'.")
+            except ApiException as create_error:
+                print(f"Failed to create ServiceAccount '{name}': {create_error}")
+                raise create_error
+        else:
+            print(f"Failed to check ServiceAccount existence: {e}")
+            raise e
+
+
+
+def create_k8s_role(role_name: str, namespace: str, rules: List[Dict[str, Any]], context_name: Optional[str] = None):
+    if not context_name:
+        context_name = get_tensorkube_cluster_context_name()
+        if not context_name:
+            raise Exception("No context found for the cluster. Please configure Tensorfuse properly.")
+
+    k8s_client = config.new_client_from_config(context=context_name)
+    rbac_api = client.RbacAuthorizationV1Api(k8s_client)
+    role = client.V1Role()
+    role.metadata = client.V1ObjectMeta(name=role_name)
+    role.rules = rules
+    try:
+        rbac_api.create_namespaced_role(namespace=namespace, body=role)
+    except ApiException as e:
+        raise e
+
+def create_role_binding(rb_name: str, rb_namespace: str, role_name: str, sa_name: str, sa_namespace: str, context_name: Optional[str] = None):
+    if not context_name:
+        context_name = get_tensorkube_cluster_context_name()
+        if not context_name:
+            raise Exception("No context found for the cluster. Please configure Tensorfuse properly.")
+
+    k8s_api_client = config.new_client_from_config(context=context_name)
+    rbac_api = client.RbacAuthorizationV1Api(k8s_api_client)
+    role_binding = client.V1RoleBinding(
+        role_ref=client.V1RoleRef(api_group="rbac.authorization.k8s.io", kind="Role", name=role_name),
+        subjects=[client.RbacV1Subject(kind="ServiceAccount", name=sa_name, namespace=sa_namespace)])
+    role_binding.metadata = client.V1ObjectMeta(name=rb_name)
+    try:
+        rbac_api.create_namespaced_role_binding(namespace=rb_namespace, body=role_binding)
+    except ApiException as e:
+        raise e
+
+
+def is_existing_scaled_job(job_name: str, namespace: str, context_name: Optional[str] = None):
+    if not context_name:
+        context_name = get_tensorkube_cluster_context_name()
+        if not context_name:
+            raise Exception("No context found for the cluster. Please configure Tensorfuse properly.")
+    k8s_api_client = config.new_client_from_config(context=context_name)
+    k8s_client = client.CustomObjectsApi(k8s_api_client)
+    try:
+        job = k8s_client.get_namespaced_custom_object(
+            group="keda.sh",
+            version="v1alpha1",
+            plural="scaledjobs",
+            name=job_name,
+            namespace=namespace
+        )
+        return True, job
+    except ApiException as e:
+        if e.status == 404:
+            return False, None
+        raise e
+
+
+def apply_domain_claim(domain: str, namespace: str):
+    k8s_api_client = config.new_client_from_config(context=get_tensorkube_cluster_context_name())
+    k8s_client = client.CustomObjectsApi(k8s_api_client)
+
+    config_yaml_path = resource_filename('tensorkube', 'configurations/build_configs/domain_claim_base_config.yaml')
+
+    with open(config_yaml_path, 'r') as f:
+        config_yaml = yaml.safe_load(f)
+
+    config_yaml['metadata']['name'] = domain
+    config_yaml['spec']['namespace'] = namespace
+
+    group = 'networking.internal.knative.dev'
+    version = 'v1alpha1'
+    plural = 'clusterdomainclaims'
+
+    try:
+        # Use cluster-scoped methods instead of namespaced ones
+        existing_claim = k8s_client.get_cluster_custom_object(
+            group=group, version=version,
+            plural=plural, name=domain
+        )
+        resource_version = existing_claim['metadata']['resourceVersion']
+        config_yaml['metadata']['resourceVersion'] = resource_version
+        k8s_client.patch_cluster_custom_object(
+            group=group, version=version,
+            plural=plural, name=domain,
+            body=config_yaml
+        )
+        click.echo(f"Updated ClusterDomainClaim for domain {domain}")
+    except ApiException as e:
+        if e.status == 404:
+            k8s_client.create_cluster_custom_object(
+                group=group, version=version,
+                plural=plural, body=config_yaml
+            )
+            click.echo(f"Created ClusterDomainClaim for domain {domain}")
+        else:
+            raise e
+        
+
+
+def patch_service_account(role_arn, service_account_name, namespace):
+    # Load the Kubernetes configuration
+    config.load_kube_config()
+
+    # Create the Kubernetes API client
+    v1 = client.CoreV1Api()
+
+    # Define the patch for the ServiceAccount
+    patch_body = {
+        "metadata": {
+            "annotations": {
+                "eks.amazonaws.com/role-arn": role_arn
+            }
+        }
+    }
+
+    try:
+        # Patch the existing ServiceAccount
+        response = v1.patch_namespaced_service_account(
+            name=service_account_name,
+            namespace=namespace,
+            body=patch_body
+        )
+        print(f"ServiceAccount '{service_account_name}' in namespace '{namespace}' patched successfully with role ARN.")
+    except client.exceptions.ApiException as e:
+        print(f"Failed to patch ServiceAccount: {e.reason}")
+
+
+def get_jobs_by_container_image(image: str, namespace: str = "keda") -> list:
+    context_name = get_tensorkube_cluster_context_name()
+    if not context_name:
+        raise Exception("Unable to get cluster context")
+    k8s_api_client = config.new_client_from_config(context=context_name)
+    k8s_client = client.CustomObjectsApi(k8s_api_client)
+
+    try:
+        scaled_jobs = k8s_client.list_namespaced_custom_object(
+            group="keda.sh",
+            version="v1alpha1",
+            plural="scaledjobs",
+            namespace=namespace
+        )
+        matching_jobs = []
+        for job in scaled_jobs.get("items", []):
+            containers = job.get("spec", {}).get("jobTargetRef", {}).get("template", {}).get("spec", {}).get(
+                "containers", [])
+            if any(container.get("image") == image for container in containers):
+                matching_jobs.append(job)
+        return matching_jobs
+
+    except ApiException as e:
+        if e.status == 404:
+            return []
+        raise e
